/*
 * I2C generische Routinen (STM32)
 * "generisch" heisst in diesem Fall nicht Plattformunabhaengigkeit, sondern
 * dass alle I2C-Implementierungen auf diesen Grundstock zurueckgreifen.
 *
 * Erzeugt: 06.04.2014
 *  Author: Nicolas
 */


#include "../src_avrcom/delay.h" // Fuer Bitbangig
#include "../src_avrcom/gpio.h"  // Fuer Bitbangig
#include "../src_misc/clock.h"
#include "i2c_generic.h"





/* Initialisierung eines der beiden I2Cs auf den Default-Pins
 * Remap wird nicht unterstuetzt */
uint_fast8_t
I2C_init(I2C_TypeDef* I2Cx)
{

	GPIO_InitTypeDef GPIO_InitStruct;
	I2C_InitTypeDef I2C_InitStruct;

	I2C_DeInit(I2Cx);

	if(I2Cx==I2C1) {

		// enable APB1 peripheral clock for I2C1, SCL and SDA pins
		RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C1,ENABLE);
		RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOB | RCC_APB2Periph_AFIO,ENABLE);

		// Setup SCL und SDA pins
		GPIO_InitStruct.GPIO_Pin   = GPIO_Pin_6 | GPIO_Pin_7;
		GPIO_InitStruct.GPIO_Mode  = GPIO_Mode_AF_OD;
		GPIO_InitStruct.GPIO_Speed = GPIO_Speed_50MHz;
		GPIO_Init(GPIOB, &GPIO_InitStruct);

		GPIO_PinRemapConfig(GPIO_Remap_I2C1,DISABLE);

		/* Enable and release I2C1 reset state */
		RCC_APB1PeriphResetCmd(RCC_APB1Periph_I2C1, ENABLE);
		RCC_APB1PeriphResetCmd(RCC_APB1Periph_I2C1, DISABLE);

	}
	else if(I2Cx==I2C2) {
		// enable APB1 peripheral clock for I2C2, SCL and SDA pins
		RCC_APB1PeriphClockCmd(RCC_APB1Periph_I2C2,ENABLE);
		RCC_APB2PeriphClockCmd(RCC_APB2Periph_GPIOB | RCC_APB2Periph_AFIO,ENABLE);

		// Setup SCL und SDA pins
		GPIO_InitStruct.GPIO_Pin   = GPIO_Pin_10 | GPIO_Pin_11;
		GPIO_InitStruct.GPIO_Mode  = GPIO_Mode_AF_OD;
		GPIO_InitStruct.GPIO_Speed = GPIO_Speed_50MHz;
		GPIO_Init(GPIOB, &GPIO_InitStruct);

		// I2C2 laesst sich nicht remappen

		/* Enable and release I2C1 reset state */
		RCC_APB1PeriphResetCmd(RCC_APB1Periph_I2C2, ENABLE);
		RCC_APB1PeriphResetCmd(RCC_APB1Periph_I2C2, DISABLE);
	}
	else {
		return i2c_off;
	}

	// configure I2Cx
	if(I2Cx==I2C1)
		I2C_InitStruct.I2C_ClockSpeed = I2C1_CLOCK;     // Hz
	else if(I2Cx==I2C2)
		I2C_InitStruct.I2C_ClockSpeed = I2C2_CLOCK;     // Hz

	I2C_InitStruct.I2C_Mode = I2C_Mode_I2C;
    #if I2C_DUTY_CYCLE
	    I2C_InitStruct.I2C_DutyCycle = I2C_DutyCycle_16_9;
    #else
	    I2C_InitStruct.I2C_DutyCycle = I2C_DutyCycle_2;
    #endif

	I2C_InitStruct.I2C_OwnAddress1 = 0x00;
	I2C_InitStruct.I2C_Ack = I2C_Ack_Disable;
	I2C_InitStruct.I2C_AcknowledgedAddress = I2C_AcknowledgedAddress_7bit;
	I2C_Init(I2Cx, &I2C_InitStruct);

	return i2c_ok;
}


/* I2C-Reset nach Phillips AN10216 I2C Manual Slide 42 ("freitakten")                   */
/* Remappte Pins werden nicht beruecksichtigt */
uint_fast8_t I2C_reset(I2C_TypeDef* I2Cx)
{
	uint8_t status, i;
	uint16_t GPIO_SCL, GPIO_SDA;
	GPIO_InitTypeDef GPIO_InitStruct;

	// Pins fuer GPIO-Operationen aussuchen
	if(I2Cx==I2C1) {
		GPIO_SCL = GPIO_Pin_6;
		GPIO_SDA = GPIO_Pin_7;
	}
	else if (I2Cx==I2C2) {
		GPIO_SCL = GPIO_Pin_10;
		GPIO_SDA = GPIO_Pin_11;
	}
	else {
		return i2cError;
	}

	// I2C abtrennen, der Rest wird durch Bitbanging mit 50 kHz erledigt
	I2C_DeInit(I2Cx);

#ifdef STM32F10x
	// Setup SCL und SDA pins
	GPIO_InitStruct.GPIO_Pin   = GPIO_SCL | GPIO_SDA;
	GPIO_InitStruct.GPIO_Mode  = GPIO_Mode_Out_OD;
	GPIO_InitStruct.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_Init(GPIOB, &GPIO_InitStruct);

#elif defined(STM32F4xx)
	// Setup SCL und SDA pins
#warning "Nie Kompiliert, nie getestet!"
	GPIO_InitStruct.GPIO_Pin   = GPIO_SCL | GPIO_SDA;
	GPIO_InitStruct.GPIO_Mode  = GPIO_Mode_Out;
	GPIO_InitStruct.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStruct.GPIO_OType = GPIO_OType_OD;
	GPIO_InitStruct.GPIO_PuPd  = GPIO_PuPd_UP;
	GPIO_Init(GPIOB, &GPIO_InitStruct);         // init GPIOB
#else
#error "No MCU defined"
#endif

	// Bus recovery
	// 1. Send 9 clock pulses on SCL line
	SetBit(GPIOB,GPIO_SCL);
	SetBit(GPIOB,GPIO_SDA);
	for(i = 0;i<9;i++) {
		delay_us(I2C_BITBANGING_DELAY_US);
		ClearBit(GPIOB,GPIO_SCL);
		delay_us(I2C_BITBANGING_DELAY_US);
		SetBit(GPIOB,GPIO_SCL);
	}

	{
		uint_least24_t timeout;
		int_fast16_t diff;
		// 2. Ask the master to keep SDA high until the "Slave Transmitter"
		// releases the SDA line to perform the ACK operation
		timeout = glu_millis +  I2C_RESETTIMEOUT;
		while( !(GPIOB->IDR & GPIO_SDA) ) {
			if(diff = timeout-glu_millis, diff < 0) {
				return i2cError;
			}
		}
	}

	// 3. Keeping SDA high during the ACK means that the "Master Receiver" does
	//    not acknowledge the previous byte receive.
	// 4. The "Slave Transmitter" then goes to an idle state

	// 5. The master then sends a STOP command initializing completely the bus
	ClearBit(GPIOB,GPIO_SCL);
	delay_us(I2C_BITBANGING_DELAY_US);
	ClearBit(GPIOB,GPIO_SDA);
	delay_us(I2C_BITBANGING_DELAY_US);
	SetBit(GPIOB,GPIO_SCL);
	delay_us(I2C_BITBANGING_DELAY_US);
	SetBit(GPIOB,GPIO_SDA);

	// Hardware I2C wieder uebernehmen lassen
	status = I2C_init(I2Cx);
	return status;
}

