/* Einfache Variante von I2C-Routinen
 * als bung zur Inbetriebnahme
 * angelehnt an:
 * http://eliaselectronics.com/stm32f4-tutorials/stm32f4-i2c-master-tutorial/
 *
 * Timeouts benoetigen externen Timer uint_least24_t glu_millis (sonst wird
 * z.B. der Scan steckenbleiben)
 *
 * Diese Implementierung mit der deutlichen Trennung der einzelnen I2C-Anteile
 * entspricht nicht komplett der Referenzimplementierung und kann bei einzelnen
 * Busteilnehmern (z.B. TMC222) bei hoeheren Taktraten zu Fehlern fuehren. 
 *
 * Ausserhalb von Prototypenszenarien sollten diese Routinen deshalb nicht mehr
 * verwendet werden. 
 *
 * 22.12.13 Nicolas
 */

/*
 * 13.04.14 Review - entspricht aktuellem Stand
 */
 
#include "misc.h"
#include "i2c_simple.h"
#include "i2c_generic.h"
#include "../../main.h"        // #define I2C_simple
#include "../src_misc/clock.h" // glu_millis


#define checkforerror(s,l) if(s) {s |= (l<<4); goto error;}



/* Busy-Status abwarten, muss vor jedem Start ausgefuehrt werden              */
uint_fast8_t I2C_waitbusy(I2C_TypeDef* I2Cx)
{
	uint_fast8_t status;
	uint_least24_t timeout;
	int_fast16_t diff;

	timeout = glu_millis + I2C_BUSYTIMEOUT;
	while(I2C_GetFlagStatus(I2Cx, I2C_FLAG_BUSY)) {
		if(diff = timeout-glu_millis, diff < 0) {
			status = i2c_busyTimeout;
			return status;
		}
	}
	status = i2c_ok;
	return status;
}


/* Start condition und R/W-Bit senden                                         *
 * I2Cx     : I2C1 oder I2C2                                                  *
 * address  : 7-Bit-Adresse linksbuendig                                      *
 * direction: I2C_Direction_Tranmitter (Master+Senden) oder                   *
 *            I2C_Direction_Receiver (Master+Empfangen)                       */
uint_fast8_t I2C_start(I2C_TypeDef* I2Cx, uint8_t address, uint8_t direction)
{
	uint_fast8_t status;
	uint_least24_t timeout;
	int_fast16_t diff;

	// START condition senden
	I2C_GenerateSTART(I2Cx, ENABLE);

	// warten auf Event 5 --> Slave has acknowledged start condition
	timeout = glu_millis + I2C_NORMALTIMEOUT;
	while(!I2C_CheckEvent(I2Cx, I2C_EVENT_MASTER_MODE_SELECT)) {
		if(diff = timeout-glu_millis, diff < 0) {
			status = i2c_startTimeout;
			return status;
		}
	}

	// Slave-Adresse + R/W-Bit senden
	I2C_Send7bitAddress(I2Cx, address, direction);

	// Warten auf Event 6 (UM0427 S. 207), je nach send/receive mit
	// unterschiedlichen Bezeichnern
	timeout = glu_millis + I2C_NORMALTIMEOUT;
	if(direction == I2C_Direction_Transmitter) {
		while(!I2C_CheckEvent(I2Cx, I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED)) {
			if (diff = timeout-glu_millis, diff < 0) {
				status = i2c_ackTimeout;
				return status;;
			}
		}
	}
	else if(direction == I2C_Direction_Receiver){
		while(!I2C_CheckEvent(I2Cx, I2C_EVENT_MASTER_RECEIVER_MODE_SELECTED)) {
			if (diff = timeout-glu_millis, diff < 0) {
				status = i2c_otherTimeout;
				return status;
			}
		}
	}

	status = i2c_ok;
	return status;
}


/* Ein Byte an Slave senden                                                   */
uint_fast8_t I2C_write(I2C_TypeDef* I2Cx, uint8_t data)
{
	uint_fast8_t status;
	uint_least24_t timeout;
	int_fast16_t diff;

	I2C_SendData(I2Cx, data);
	// wait for I2C1 EV8_2 --> byte has been transmitted
	timeout = glu_millis + I2C_NORMALTIMEOUT;
	while(!I2C_CheckEvent(I2Cx, I2C_EVENT_MASTER_BYTE_TRANSMITTED) ) {
		if(diff = timeout-glu_millis, diff < 0) {
			status =  i2c_ackTimeout;
			return status;
		}
	}
	status = i2c_ok;
	return status;
}


/* Ein Byte vom Slave lesen und mit ACK besttigen = ein weiteres Byte        *
 * anfordern                                                                  */
uint_fast8_t I2C_read_ack(I2C_TypeDef* I2Cx, uint8_t *data)
{
	uint_fast8_t status;
	uint_least24_t timeout;
	int_fast16_t diff;

	// enable acknowledge of recieved data
	I2C_AcknowledgeConfig(I2Cx, ENABLE);
	// wait until one byte has been received
	timeout = glu_millis + I2C_NORMALTIMEOUT;
	while( !I2C_CheckEvent(I2Cx, I2C_EVENT_MASTER_BYTE_RECEIVED) ) {
		if(diff = timeout-glu_millis, diff < 0) {
			status = i2c_ackTimeout;
			return status;
		}
	}

	// I2C data register auslesen
	*data = I2C_ReceiveData(I2Cx);
	return i2c_ok;
}


/* Ein Byte vom Slave lesen und mit NACK besttigen = kein weiteres Byte      *
 * anfordern                                                                  */
uint_fast8_t I2C_read_nack(I2C_TypeDef* I2Cx, uint8_t *data)
{
	uint_fast8_t status;
	uint_least24_t timeout;
	int_fast16_t diff;

	// disabe acknowledge of received data
	// nack also generates stop condition after last byte received
	// see reference manual for more info
	I2C_AcknowledgeConfig(I2Cx, DISABLE);
	I2C_GenerateSTOP(I2Cx, ENABLE);
	// wait until one byte has been received
	timeout = glu_millis + I2C_NORMALTIMEOUT;
	while( !I2C_CheckEvent(I2Cx, I2C_EVENT_MASTER_BYTE_RECEIVED) ) {
		if(diff = timeout-glu_millis, diff < 0) {
			status = i2c_nAckTimeout;
			return status;
		}
	}

	// read data from I2C data register and return data byte
	*data = I2C_ReceiveData(I2Cx);
	status = i2c_ok;
	return status;
}


/* Sendet STOP condition und gibt den Bus frei */
uint_fast8_t I2C_stop(I2C_TypeDef* I2Cx)
{
	// Send I2C1 STOP Condition
	I2C_GenerateSTOP(I2Cx, ENABLE);
	return i2c_stop;
}


#ifdef I2C_SIMPLE

/* Ganzen Puffer ueber I2C lesen                                                        */
/* I2Cx     : I2C1 oder I2C2 (bei AVR egal)
 * slaveaddr: I2C-Slave-Adresse
 * *pBuffer : Puffer passender Groesse
 * nBuffer  : Puffergroesse
 *
 * return   : Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
 *            und STOP-Bedingung erzeugt werden
 *
 * wenn pBuffer == NULL werden nBuffer Bytes vom Bus gelesen und verworfen.
 * wenn nBuffer == NULL wird nur Stopbedingung gesendet */
uint_fast8_t
I2C_Receivebuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
		uint8_t* pBuffer, uint8_t nBuffer)
{
	uint8_t status, i, byte;

	if (nBuffer==0) {
		I2C_stop(I2Cx);
		return 0;
	}

	status = I2C_start(I2Cx,slaveaddr, I2C_Direction_Receiver);
	checkforerror(status,5);

	for(i=0;i<nBuffer-1;i++) {
		status = I2C_read_ack(I2Cx,&byte);
		checkforerror(status,6);
		if(pBuffer) {
			pBuffer[i] = byte;
		}
	}

	status = I2C_read_nack(I2Cx,&byte);
	checkforerror(status,7);
	if(pBuffer) {
		pBuffer[nBuffer-1] = byte;
	}

	error:
	I2C_stop(I2Cx);
	return status;
}


/* Ganzen Puffer ueber I2C senden                                                       */
/* I2Cx     : I2C1 oder I2C2 (bei AVR egal)
 * slaveaddr: I2C-Slave-Adresse
 * *pBuffer : Puffer passender Groesse
 * nBuffer  : Puffergroesse
 * stop     : 1: Stop-Bedingung wird gesendet (normales Senden an Slave)
 *            0: Stop-Bedingung wird nicht gesendet (Antwort von Slave
 *            vorbereiten)
 *
 * return   : Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
 *            und STOP-Bedingung erzeugt werden  */
uint_fast8_t
I2C_Sendbuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
		uint8_t* pBuffer, uint8_t nBuffer, enum stopbit_e stopbit )
{
	  uint_fast8_t status, i;

	  status = I2C_waitbusy(I2Cx);
	  checkforerror(status,1);

	  status = I2C_start(I2Cx,slaveaddr, I2C_Direction_Transmitter);
	  checkforerror(status,2);

	  for(i=0;i<nBuffer;i++) {
	    status = I2C_write(I2Cx,pBuffer[i]);
	    checkforerror(status,4);
	  }

	  switch (stopbit) {
	  case stop:
		  I2C_stop(I2Cx);
		  break;
	  case noStop:
		  ;
	  }

	  error:
	  return status;
}


/* I2C-Adresse kurz anpingen um zu sehen, ob Geraet vorhanden                           */
/*
* return: i2c_ok          : Device vorhanden und hat mit ACK bestaetigt
 *         i2c_busyTimeout : Bus nicht frei
 *         i2c_otherTimeout: Unerwarteter Fehler
 *         i2c_ackTimeout  : NACK, Device nicht vorhanden */
uint_fast8_t
I2C_ping(I2C_TypeDef *I2Cx, uint8_t slaveaddr)
{
	uint_fast8_t status;

	// i2c_ok oder i2c_busyTimeout
	status = I2C_waitbusy(I2Cx);
	if(status) {
	    status |= (2<<4);
	    goto error;
	}

	// Startbedingung
	// i2c_ok           : ACK empfangen, Teilnehmer vorhanden
	// i2c_ackTimeout   : kein ACK empfangen, keine Antwort vom Teilnehmer
	// i2c_otherTimeout : Start-Bedingung konnte nicht gesendet werden.
	//                    Senderfehler.
	status = I2C_start(I2Cx,slaveaddr,I2C_Direction_Transmitter);
	I2C_stop(I2Cx);

	error:
	return status;
}

#endif
