/*
 * I2C (polling-Modus) fuer STM32F10x
 * nach Application Note AN2824
 * Beruecksichtigt Hardwarefehler im I2C-Modul
 *
 * Timeouts benoetigen externen Timer uint_least24_t glu_millis
 *
 * Die CMSIS-Structs koennen wegen des Hardwarefehlers nicht verwendet werden,
 * deshalb muessen die Hardware-Adressen direkt hinterlegt werden.
 *
 * Erzeugt: 31.03.2014
 *  Author: Nicolas
 */



#ifndef STM32F10x
  #error "only STM32F10x supported"
#endif

#include "stm32f10x_gpio.h"
#include "stm32f10x_rcc.h"
#include "stm32f10x_i2c.h"

#include "../src_misc/clock.h" // glu_millis
#include "../../main.h"
#include "i2c_generic.h"
#include "i2c_stm32f10x.h"

#ifdef I2C_DEVICE_SPECIFIC




// Globale Variablen fuer I2C-Status
// Fixme: unbenoetigte Variablen
I2C_TypeDef * I2C_Module;
volatile uint8_t deviceAddress;
volatile uint8_t dataByte1;
volatile uint8_t dataByte0;
volatile uint8_t receivedDataByte1;
volatile uint8_t receivedDataByte0;
volatile uint8_t i2cDirectionWrite;
volatile uint8_t i2cByteCounter;
volatile uint8_t i2cBusyFlag;




/* ************************************************************************** */
/* Hardwareadressen STM32F10x                                                 */

/* I2C START mask */
#define CR1_START_Set           ((uint16_t)0x0100)
#define CR1_START_Reset         ((uint16_t)0xFEFF)

/* I2C STOP mask */
#define CR1_STOP_Set            ((uint16_t)0x0200)
#define CR1_STOP_Reset          ((uint16_t)0xFDFF)

/* I2C ACK mask */
#define CR1_ACK_Set             ((uint16_t)0x0400)
#define CR1_ACK_Reset           ((uint16_t)0xFBFF)

#define CR1_POS_Set             ((uint16_t)0x0800)
#define CR1_POS_Reset           ((uint16_t)0xF7FF)

#define I2C_IT_BUF              ((uint16_t)0x0400)
#define I2C_IT_EVT              ((uint16_t)0x0200)
#define I2C_IT_ERR              ((uint16_t)0x0100)




/* Ganzen Puffer ueber I2C senden                                             */
/* I2Cx     : I2C1 oder I2C2 (bei AVR egal)
 * slaveaddr: I2C-Slave-Adresse
 * *pBuffer : Puffer passender Groesse
 * nBuffer  : Puffergroesse
 * stop     : 1: Stop-Bedingung wird gesendet (normales Senden an Slave)
 *            0: Stop-Bedingung wird nicht gesendet (Antwort von Slave
 *            vorbereiten)
 *
 * return   : Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
 *            und STOP-Bedingung erzeugt werden
 *
 * Timeout benoetigt externen Timer glu_millis
 */
uint_fast8_t
I2C_Sendbuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
		uint8_t* pBuffer, uint8_t nByte, enum stopbit_e stopbit)
{
	uint_least24_t timeout;
	int_fast16_t diff;

	/* Error-Interrupt einschalten (RM0008 S.747)                             */
    #ifndef DISABLE_I2C_ERROR_IRQ
	    I2Cx->CR2 |= I2C_IT_ERR;
    #else
        #warning Fehler-interrupt aus
    #endif

	/* Send START condition */
	I2Cx->CR1 |= CR1_START_Set;
	/* Wait until SB flag is set: EV5 */
    timeout = glu_millis + I2C_NORMALTIMEOUT;
	while ((I2Cx->SR1&0x0001) != 0x0001) {
		if (diff = timeout-glu_millis, diff < 0)
			return i2cError;
	}

	/* Send slave address */
	/* Reset the address bit0 for write*/
	slaveaddr &= ~1U;
	I2Cx->DR = slaveaddr;
	/* Wait until ADDR is set: EV6 */
    timeout = glu_millis + I2C_NORMALTIMEOUT;
	while ((I2Cx->SR1 &0x0002) != 0x0002) {
		if (diff = timeout-glu_millis, diff < 0)
			return i2cError;
	}


	I2Cx->SR2;           // Clear ADDR flag by reading SR2 register
	I2Cx->DR = *pBuffer; // Write the first data in DR register (EV8_1)
	pBuffer++;           // Increment
	nByte--;
	while (nByte--) {
		/* Poll on BTF to receive data because in polling mode we can not
		 * guarantee the EV8 software sequence is managed before the current
		 * byte transfer completes */
		while ((I2Cx->SR1 & 0x00004) != 0x000004);
		/* Send the current byte */
		I2Cx->DR = *pBuffer;
		/* Point to the next byte to be written */
		pBuffer++;
	}

	/* EV8_2: Wait until BTF is set before programming the STOP */
	while ((I2Cx->SR1 & 0x00004) != 0x000004);

	if (stopbit) {
	    I2Cx->CR1 |= CR1_STOP_Set; // Send STOP Condition

	    /* Make sure that the STOP bit is cleared by Hardware */
	    while ((I2Cx->CR1&0x200) == 0x200);
	}

    return 0;
}


/* Ganzen Puffer ueber I2C lesen                                              */
/* I2Cx     : I2C1 oder I2C2 (bei AVR egal)
 * slaveaddr: I2C-Slave-Adresse
 * *pBuffer : Puffer passender Groesse
 * nBuffer  : Puffergroesse
 *
 * return   : Fehlerbedingung; wenn != 0, muss I2C vom Aufrufer zurueckgesetzt
 *            und STOP-Bedingung erzeugt werden
 *
 * wenn pBuffer == NULL werden nBuffer Bytes vom Bus gelesen und verworfen.
 * wenn nBuffer == NULL wird nur Stopbedingung gesendet
 */
uint_fast8_t
I2C_Receivebuffer(I2C_TypeDef *I2Cx, uint8_t slaveaddr,
		uint8_t* pBuffer, uint8_t nByte)
{
    /* Je nachdem ob ein, zwei oder mehrere Bytes empfangen werden muss das
     * Verfahren angepasst werden.
	 */

	uint_least24_t timeout;
	int_fast16_t diff;

    #ifndef DISABLE_I2C_ERROR_IRQ
        I2Cx->CR2 |= I2C_IT_ERR; // Enable I2C error interrupts
    #else
        #warning "error-Interrupt aus"
    #endif

    if (nByte == 1) {


    	I2Cx->CR1 |= CR1_START_Set; // Send START condition

    	/* Wait until SB flag is set: EV5  */
        timeout = glu_millis + I2C_NORMALTIMEOUT;
    	while ((I2Cx->SR1&0x0001) != 0x0001) {
    		if (diff = timeout-glu_millis, diff < 0)
    			return i2cError;
    	}

    	slaveaddr |= 1U;             // Set the address bit0 for read
    	I2Cx->DR = slaveaddr;        // Send the slave address

    	/* Wait until ADDR is set: EV6_3, then program ACK = 0, clear ADDR
           and program the STOP just after ADDR is cleared. The EV6_3
           software sequence must complete before the current byte end of
           transfer.*/

    	/* Wait until ADDR is set */
    	timeout = glu_millis + I2C_NORMALTIMEOUT;
    	while ((I2Cx->SR1&0x0002) != 0x0002) {
    		if (diff = timeout-glu_millis, diff < 0)
    			return i2cError;
    	}

    	I2Cx->CR1 &= CR1_ACK_Reset; // Clear ACK bit

    	/* Disable all active IRQs around ADDR clearing and STOP programming
    	 * because the  EV6_3 software sequence must complete before the current
    	 * byte end of transfer */
    	__disable_irq();
    	I2Cx->SR2;                 // Clear ADDR flag
    	I2Cx->CR1 |= CR1_STOP_Set; // Program the STOP
    	__enable_irq();

    	/* Wait until a data is received in DR register (RXNE = 1) EV7 */
    	while ((I2Cx->SR1 & 0x00040) != 0x000040);

    	*pBuffer = I2Cx->DR; // Read the data

    	/* Make sure that the STOP bit is cleared by Hardware before CR1 write
    	 * access */
    	while ((I2Cx->CR1&0x200) == 0x200);
    	/* Enable Acknowledgement to be ready for another reception */
    	I2Cx->CR1 |= CR1_ACK_Set;
    } // if nByte==1
    else if (nByte==2) {

    	I2Cx->CR1 |= CR1_POS_Set; // Set POS bit
    	I2Cx->CR1 |= CR1_START_Set; // Send START condition

    	/* Wait until SB flag is set: EV5 */
    	timeout = glu_millis + I2C_NORMALTIMEOUT;
    	while ((I2Cx->SR1&0x0001) != 0x0001) {
    		if (diff = timeout-glu_millis, diff < 0)
    			return i2cError;
    	}


    	/* Send the slave address */
    	slaveaddr |= 1U;   	            // Set the address bit0 for read
    	I2Cx->DR = slaveaddr;           // Send slave address

    	/* Wait until ADDR is set: EV6 */
    	timeout = glu_millis + I2C_NORMALTIMEOUT;
    	while ((I2Cx->SR1&0x0002) != 0x0002) {
    		if (diff = timeout-glu_millis, diff < 0)
    			return i2cError;
    	}

    	/* EV6_1: The acknowledge disable should be done just after EV6, that
    	 * is after ADDR is cleared, so disable all active IRQs around ADDR
    	 * clearing and ACK clearing */
    	__disable_irq();
    	I2Cx->SR2;                  // ADDR by reading SR2 register
    	I2Cx->CR1 &= CR1_ACK_Reset; // Clear ACK
    	__enable_irq();

    	/* Wait until BTF is set */
    	while ((I2Cx->SR1 & 0x00004) != 0x000004);

    	/* Disable IRQs around STOP programming and data reading because of the
    	 * limitation ?*/
    	__disable_irq();
    	I2C_GenerateSTOP(I2Cx, ENABLE); // Program the STOP
    	*pBuffer = I2Cx->DR;     	    // Read first data
    	__enable_irq();

    	pBuffer++;
    	*pBuffer = I2Cx->DR;   // Read second data

    	/* Make sure that the STOP bit is cleared by Hardware before CR1 write
    	 * access */
    	while ((I2Cx->CR1&0x200) == 0x200);

    	/* Enable Acknowledgement to be ready for another reception */
    	I2Cx->CR1  |= CR1_ACK_Set;
    	/* Clear POS bit */
    	I2Cx->CR1  &= CR1_POS_Reset;

    } // nByte==2
    else {

    	I2Cx->CR1 |= CR1_START_Set;     	// Send START condition

    	/* Wait until SB flag is set: EV5 */
    	timeout = glu_millis + I2C_NORMALTIMEOUT;
    	while ((I2Cx->SR1&0x0001) != 0x0001) {
    		if (diff = timeout-glu_millis, diff < 0)
    			return i2cError;
    	}

    	slaveaddr |= 1U;   	            // Set the address bit0 for read
    	I2Cx->DR = slaveaddr;           // Send slave address


    	/* Wait until ADDR is set: EV6 */
    	timeout = glu_millis + I2C_NORMALTIMEOUT;
    	while ((I2Cx->SR1&0x0002) != 0x0002) {
    		if (diff = timeout-glu_millis, diff < 0)
    			return i2cError;
    	}

    	I2Cx->SR2;                  // Clear ADDR by reading SR2 status register
    	/* While there is data to be read */
    	while (nByte)
    	{
    		/* Receive bytes from first byte until byte N-3 */
    		if (nByte != 3) {
    			/* Poll on BTF to receive data because in polling mode we can
    			 * not  guarantee the EV7 software sequence is managed before
    			 * the current byte transfer completes */
    			while ((I2Cx->SR1 & 0x00004) != 0x000004);

    			*pBuffer = I2Cx->DR; // Read Data
    			pBuffer++;
    			nByte--;
    		}

    		/* it remains to read three data: data N-2, data N-1, Data N */
    		if (nByte == 3)
    		{

    			/* Wait until BTF is set: Data N-2 in DR and data N -1 in shift
    			 * register */
    			while ((I2Cx->SR1 & 0x00004) != 0x000004);

    			I2Cx->CR1 &= CR1_ACK_Reset; // Clear ACK

    			/* Disable IRQs around data reading and STOP programming
    			 * because of the limitation ? */
    			__disable_irq();
    			*pBuffer = I2Cx->DR;       // Read Data N-2
    			pBuffer++;
    			I2Cx->CR1 |= CR1_STOP_Set; // Program the STOP
    			*pBuffer = I2Cx->DR;       // Read DataN-1
     			__enable_irq();

     			pBuffer++;
    			/* Wait until RXNE is set (DR contains the last data) */
    			while ((I2Cx->SR1 & 0x00040) != 0x000040);

    			*pBuffer = I2Cx->DR; // Read DataN
    			nByte = 0;
    		}
    	}

    	/* Make sure that the STOP bit is cleared by Hardware before CR1 write
    	 * access */
    	while ((I2Cx->CR1&0x200) == 0x200);
    	/* Enable Acknowledgement to be ready for another reception */
    	I2Cx->CR1 |= CR1_ACK_Set;
    }

    return i2c_ok;
}


/* I2C-Adresse kurz anpingen um zu sehen, ob Geraet vorhanden                           */
/*
 * return: i2c_ok          : Device vorhanden und hat mit ACK bestaetigt
 *         i2c_ackTimeout  : NACK, Device nicht vorhanden
 *         i2c_busyTimeout : Bus nicht frei
 *         i2c_otherTimeout: Unerwarteter Fehler
 *
 * Der Aufrufer ist dafuer verantwortlich, bei einem anderen Rueckgabewert als
 * "i2c_okW oder "i2c_ackTimeout" den Bus zurueckzusetzen.
 *
 * Fuer diese Funktion gibt es keine Referenzimplementierung.
 */
uint_fast8_t
I2C_ping(I2C_TypeDef *I2Cx, uint8_t slaveaddr)
{
	uint_fast8_t status = i2c_ok;
	uint_least24_t timeout;
	int_fast16_t diff;

	/* Error-Interrupt einschalten (RM0008 S.747)                             */
    #ifndef DISABLE_I2C_ERROR_IRQ
	    I2Cx->CR2 |= I2C_IT_ERR;
    #else
        #warning Fehler-interrupt aus
    #endif


	/* Send START condition */
	I2Cx->CR1 |= CR1_START_Set;
	/* Wait until SB flag is set: EV5 */
	timeout = glu_millis + I2C_SCANTIMEOUT;
	while ((I2Cx->SR1&0x0001) != 0x0001) {
		if (diff = timeout-glu_millis, diff < 0) {
			// Dieser Timeout ist ein Fehler - STOP-Bedingung wird nicht erzeugt
			status = i2c_otherTimeout;
			return status;
		}
	}

	/* Send slave address */
	/* Reset the address bit0 for write*/
	slaveaddr &= ~1U;
	I2Cx->DR = slaveaddr;
	/* Wait until ADDR is set: EV6 */
	timeout = glu_millis + I2C_SCANTIMEOUT;
	while ((I2Cx->SR1 &0x0002) != 0x0002) {
		if (diff = timeout-glu_millis, diff < 0) {
			status = i2c_ackTimeout;
		    break;
		}
	}


	I2Cx->SR2;                 // Clear ADDR flag by reading SR2 register

	I2Cx->CR1 |= CR1_STOP_Set; // Send STOP Condition

	/* Make sure that the STOP bit is cleared by Hardware */
	while ((I2Cx->CR1&0x200) == 0x200);

	return status;
}


#endif
